""" Acessors for XML """
__author__ = 'Alistair Broomhead'

from xml.dom.minidom import NodeList
from slacken.dicttypes import AttrDict


class XMLAccessor(object):
    """ Wraps a dom element to allow attribute-style access """

    @classmethod
    def from_string(cls, text):
        """ Constructor from string """
        from xml.dom.minidom import parseString
        return cls(parseString(text))

    def to_dict(self, base_name='xml_accessor'):
        """
        :param base_name: root dictionary key
        """

        assert isinstance(self, XMLAccessor)

        def _set_tag(d, tag, data):
            d[tag] = data
            return d

        def _r(node, d, tag):
            try:
                return _set_tag(d, tag, node.data)
            except AttributeError:
                pass
            _d, _l = {}, []
            for n in node.childNodes:
                try:
                    _r(XMLAccessor(n), _d, n.tagName)
                except AttributeError:
                    _d_ = {}
                    _r(XMLAccessor(n), _d_, '')
                    _l.append(_d_[''])
            if _d:
                return _set_tag(d, tag, _d)
            if _l:
                if len(_l) == 1:
                    return _set_tag(d, tag, _l[0])
                return _set_tag(d, tag, _l)
            return _set_tag(d, tag, None)

        d = _r(self, {}, base_name)[base_name]
        return AttrDict(d)

    def __repr__(self):
        if isinstance(self.dom, NodeList):
            return "<XMLAccessor %r>" % [x.nodeName for x in self.dom]
        else:
            return "<XMLAccessor %r>" % self.dom.nodeName

    def __init__(self, dom):
        if isinstance(dom, NodeList) and len(dom) == 1:
            dom = dom[0]
        self.dom = dom

    def __getattr__(self, item):
        if isinstance(self.dom, NodeList):
            for node in self.dom:
                try:
                    return getattr(XMLAccessor(node), item)
                except BaseException:
                    pass
            from xml.dom import NotFoundErr

            raise NotFoundErr(item)
        else:
            try:
                node = self.dom.getElementsByTagName(item)
                if not node:
                    attrs = self.dom._attrs
                    return AttrDict(attrs)[item].nodeValue
                if isinstance(node, NodeList) and len(node) == 1:
                    node = node[0]
                return XMLAccessor(node)
            except BaseException:
                return getattr(self.dom, item)



"""
Tools for working with restful apis
"""
__author__ = 'Alistair Broomhead'
from slacken.dicttypes import AttrDict
from slacken.rest_access import RESTaccess
from slacken.xml_accessor import XMLAccessor



"""
Syntactic sugar
"""
__author__ = 'Alistair Broomhead'


def _accessor_factory(dict_):

    class AttrDict(object):
        """
        Accessor object to get dictionary keys by attribute access - can
        make for easier to read syntax (Javascript style)
        """
        @staticmethod
        def _as_dict():
            return dict_

        def __getattr__(self, item):
            if item in dict_.keys():
                x = dict_[item]
                if isinstance(x, dict):
                    return _accessor_factory(x)()
                return x

        def __setattr__(self, key, value):
            dict_[key] = value

        def __getitem__(self, item):
            if item in dict_.keys():
                x = dict_[item]
                if isinstance(x, dict):
                    return _accessor_factory(x)()
                return x

        def __setitem__(self, key, value):
            dict_[key] = value

        def __delattr__(self, item):
            if item in dict_.keys():
                del dict_[item]

        def __repr__(self):
            return repr(dict_)

        def __str__(self):
            return str(dict_)

        def __iter__(self):
            return iter(dict_)

    return AttrDict


def AttrDict(dict_=None):
    """
    Accessor object to get dictionary keys  by attribute access
    :param dict_: existing dictionary (if None given, creates a new one)
    """
    if dict_ is None:
        dict_ = {}
    cls = _accessor_factory(dict_)
    return cls()



""" Accessors for REST endpoints """
from urllib2 import HTTPError

from slacken.xml_accessor import XMLAccessor

__author__ = 'Alistair Broomhead'


class RESTaccess(object):
    """
    An accessor for REST endpoints

    rest_hub should be something like 'http://www.integration.moshi/services/rest'
    """
    rest_hub = ''

    @staticmethod
    def _get_raw(url, params=None):
        from urllib2 import build_opener, Request
        if params is not None:
            from urllib import urlencode
            params = urlencode(params)

        opener = build_opener()
        try:
            return opener.open(Request(url), params)
        except HTTPError:
            return opener.open(Request(url + '/'), params)

    @staticmethod
    def _parse_json(raw):
        from json import load
        return load(raw)

    @staticmethod
    def _get_json(url, params=None):
        return RESTaccess._parse_json(
            RESTaccess._get_raw(url, params)
        )

    @staticmethod
    def _parse_xml(raw):
        from xml.dom.minidom import parse
        dom = parse(raw)
        return XMLAccessor(dom)

    @staticmethod
    def _get_xml(url, params=None):
        return RESTaccess._parse_xml(
            RESTaccess._get_raw(url, params)
        )

    def __init__(self, rest_hub):
        self.rest_hub = rest_hub

    def __repr__(self):
        return 'RESTaccess(%r)' % self.rest_hub

    def url(self, endpoint):
        """ Gives the full url of the given endpoint """
        return '/'.join(
            (self.rest_hub.rstrip('/'), endpoint.lstrip('/'))).rstrip('/')

    def __call__(self, endpoint, params=None):
        """
        GETs the enpoint unless params is passed, in which case it POSTs params
        """
        url_ = self.url(endpoint)
        content = self._get_raw(url_, params)
        subtype = content.headers.subtype.lower().strip()
        assert isinstance(subtype, str)

        if subtype == 'json':
            return self._parse_json(content)
        elif subtype == 'xml':
            return self._parse_xml(content)
        else:
            return content



from setuptools import setup, find_packages
from os import path


def open_(fname):
    return open(path.join(path.dirname(__file__), fname))


def read(fname):
    return open_(fname).read()


def readlines(fname):
    return open_(fname).readlines()


def read_json(fname):
    from json import load
    return load(open_(fname))


def read_version():
    return '.'.join([str(i) for i in
                     read_json("src/slacken/version.txt")])


setup(
    name='slacken',
    package_data={'': ['*.txt']},
    author='Alistair Broomhead',
    version=read_version(),
    author_email='alistair.broomhead@gmail.com',
    description='Tools for working with rest',
    license='MIT',
    url='https://github.com/alistair-broomhead/slacken',
    download_url='https://github.com/alistair-broomhead/slacken/zipball/master',
    long_description=read('README.txt'),
    packages=find_packages('src'),
    package_dir={'': 'src'},
    #install_requires=['xml', 'urllib2']
    install_requires=readlines("requirements.txt")
)
